"""
Contains a `ProcessRequest` class to process and get a response
from a given route and requested method. Use to process client requests.
"""

import json
from typing import List, Union

import http_pyparser

from .errors import Error, default_errors
from .route import Route
from .server import Client


class RequestData:
    def __init__(
        self,
        parsed_http: http_pyparser.parser.HTTPData,
        remote_addr: tuple
    ):
        self.real_path = parsed_http.real_path

        self.path = parsed_http.path
        self.method = parsed_http.method
        self.version = parsed_http.version
        
        self.host = parsed_http.host
        self.user_agent = parsed_http.user_agent
        self.accept = parsed_http.accept

        self.body = parsed_http.body
        self.headers = parsed_http.headers
        self.cookies = parsed_http.cookies
        self.query = parsed_http.query

        self.remote_addr = remote_addr
        self.parameters = {}

    def json(self) -> Union[None, dict]:
        """Return body as JSON.

        If None is returned, it means that the request
        does not have a body. This method does not handle exceptions,
        decoding errors will be thrown by the `json` module.

        :return: Body as JSON format or None
        :rtype: Union[None, dict]
        """

        if self.body:
            data = json.loads(self.body)
        else:
            data = None

        return data

    def __repr__(self) -> str:
        return (f'RequestData(real_path="{self.real_path}", path="{self.path}", method="{self.method}", '
                f'version="{self.version}", host="{self.host}", user_agent="{self.user_agent}", '
                f'accept="{self.accept}", body={self.body}, headers={self.headers}, cookies={self.cookies}, '
                f'query={self.query}, remote_addr={self.remote_addr}, parameters={self.parameters})')


class RequestProcessed:
    def __init__(self, route: Union[Route, Error], request: RequestData) -> None:
        self.route = route
        self.request = request

        if isinstance(route, Route):
            self.type = 'route'
        elif isinstance(route, Error):
            self.type = 'error'

    def get_response(self, use_globals: bool = False) -> http_pyparser.Response:
        if self.type == 'route':
            response = self.route.get_route_response(self.request, use_globals=use_globals)
        elif self.type == 'error':
            response = self.route.get_callback_response(self.request)

        return response


class ProcessRequest:
    def __init__(self, routes: List[Route], errors_callback: List[Error] = []) -> None:
        self._routes = routes
        self._errors_callback = errors_callback
        self._errors_callback.extend(default_errors)

    def _get_route_by_path(self, path: str) -> Union[Route, None]:
        for route in self._routes:
            if route.match_route(path):
                return route

    def process(self, client: Client) -> RequestProcessed:
        """Processes raw HTTP data and creates a `ProcessedRequest`
        object. This object has route information with a `Route`
        object and the request that was made by the client, with
        the `RequestData` object.

        All manipulation of these data are done as you wish.

        :param client: A `Client` instance
        :type client: Client
        :return: Return a RequestProcessed object
        :rtype: RequestProcessed
        """

        # get client request
        message = client.get_message()

        if not message:
            client.destroy()
        else:
            http_parser = http_pyparser.parser.HTTPParser()
            parsed_http = http_parser.parser(message)
            remote_addr = client.get_address()

            route = self._get_route_by_path(parsed_http.path)
            request = RequestData(parsed_http, remote_addr)

            if route:
                request.parameters = route.get_parameters(parsed_http.path)

                if route.accept_method(request.method):
                    processed_request = RequestProcessed(route, request)
                else:
                    for error_handle in self._errors_callback:
                        if error_handle.match_status_code(405):
                            processed_request = RequestProcessed(error_handle, request)
                            break
            else:
                for error_handle in self._errors_callback:
                    if error_handle.match_status_code(404):
                        processed_request = RequestProcessed(error_handle, request)
                        break

            return processed_request



from datetime import datetime

from http_pyparser import Response
from .request import RequestData


def log_request(response: Response, request: RequestData) -> None:
    time = datetime.now()

    path = request.real_path
    status = response.status
    method = request.method
    host = request.remote_addr[0]

    print(f'[\033[33m{time}\033[m]', end='')
    print(f'[{host}]', end=' ', flush=True)

    if 100 <= status < 200 or 300 <= status < 400:
        print(f'\033[33m{status}\033[m', flush=True, end=' ')
    elif 200 <= status < 300:
        print(f'\033[32m{status}\033[m', flush=True, end=' ')
    elif 400 <= status <= 500:
        print(f'\033[31m{status}\033[m', flush=True, end=' ')

    print(f'{path} ({method})')



from types import FunctionType

from .exceptions import InvalidResponseError

from http_pyparser import response
from http_pyparser.parser import HTTPData


class Error:
    def __init__(self, callback: FunctionType, status_code: int) -> None:
        self._callback = callback
        self._status_code = status_code

    def match_status_code(self, status_code: int) -> bool:
        return self._status_code == status_code

    def get_callback_response(self, request: HTTPData) -> response.Response:
        """Gets the response from the error handler.

        The `request` argument is only passed as
        an argument to the callback function, if it
        requests, so that it can get data from the request.

        :param request: Request data
        :type request: RequestData
        :raises InvalidResponseError: If the route returns None,
        or a boolean value.
        :return: Callback response in Response object;
        :rtype: http_pyparser.parser.HTTPData
        """

        try:
            callback_response = self._callback.__call__(request)
        except TypeError:
            callback_response = self._callback.__call__()

        if not callback_response:
            raise InvalidResponseError(f'Error callback to "{self._status_code}" '
                                        'status code returned a invalid response')
        else:
            if isinstance(callback_response, tuple):
                # getting body and status of response
                # in use cases of: return "Hello", 200.
                body, status = callback_response
                final_response = response.Response(body, status=status)
            elif isinstance(callback_response, response.Response):
                final_response = callback_response
            else:
                final_response = response.Response(callback_response)

            return final_response


def not_found_error_404():
    return '<h1>404: Not Found</h1>', 404


def method_not_allowed_405():
    return '<h1><b>405</b>: Method Not Allowed</h1>', 405


default_errors = [
    Error(not_found_error_404, 404),
    Error(method_not_allowed_405, 405)
]
